@@ -18,11 +18,9 @@ void message(int level, char *format, ...)
 #else
 ;
 #endif
-#if 0
-void log(int fd, char *format, ...);
-#endif
 void logSetErrorFile(FILE * f);
 void logSetMessageFile(FILE * f);
+void logToSyslog(int enable);
 void logSetFlags(int flags);
 void logClearFlags(int flags);
 void logSetLevel(int level);
